import os
import matplotlib
matplotlib.use('Agg')
import numpy as np
import pandas as pd
import scanpy.api as sc

try:
    import tensorflow as tf
except ImportError:
    raise ImportError('DESC requires tensorflow  Please follow instructions'
                      ' at https://www.tensorflow.org/install/ to install'
                      ' it.')
def paul_test(n_top_gene=100):
    adata=sc.read("data/paul15/paul15.h5ad")
    sc.pp.filter_cells(adata,min_genes=10)
    sc.pp.normalize_per_cell(adata,counts_per_cell_after=1e4)
    sc.pp.filter_genes(adata,min_cells=20)
    sc.pp.filter_genes_dispersion(adata,n_top_genes=1000)
    sc.pp.log1p(adata)
    sc.pp.scale(adata,zero_center=True,max_value=False)
    return adata


def test_train():
    #simple filter
    adata=paul_test()
    from train import train
    
    adata=train(adata,louvain_resolution=0.4,use_GPU=True)
     
    #adata=desc(x,dims,louvain_resolution=0.4,use_GPU=False)
    
    #adata=desc(x,dims,louvain_resolution='0.4,0.6,0.8',use_GPU=True)

    #adata=desc(x,dims,louvain_resolution=[0.4,0.5,0.8],use_GPU=True)

test_train()



import os
import numpy as np


def write_desc_result(data, save_dir='tmp_result', delimiter=","):

    os.makedirs(save_dir, exist_ok=True)
    assert hasattr(
        data.obsm, 'embedded'), 'The desc result not found, run desc first'

    if not os.path.isfile(save_dir + '/prob_matrix.csv'):
        np.savetxt(save_dir + '/prob_matrix.csv',
                   data.obsm['prob'], delimiter=delimiter)
    else:
        print('The ' + save_dir + '/prob_matrix.csv has already exist!')

    if not os.path.isfile(save_dir + '/cluster_ident.cvs'):
        np.savetxt(save_dir + '/cluster_ident.csv',
                   data.obs.iloc[:, [0, 2]], delimiter=delimiter, fmt="%s")
    else:
        print('The ' + save_dir + '/cluster_ident.csv has already exist!')

    if not os.path.isfile(save_dir + '/embedded.csv'):
        np.savetxt(save_dir + '/embedded.cvs',
                   data.obsm['embedded'], delimiter=delimiter)
    else:
        print('The ' + save_dir + '/embedded.csv has already exist!')

    return None



# Copyright 2018 Xiangjie Li
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

#from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')
import os,math
from time import time
import random
import numpy as np
import pandas as pd
import tensorflow as tf
from anndata import AnnData
import scanpy.api as sc
from keras import backend as K
from scipy.sparse import issparse
import keras
try:
    from .network import *
except:
    from network import *
    

#check use gpu
#if len(K.tensorflow_backend._get_available_gpus()):
#if tf.test.is_gpu_available():
#config = tf.ConfigProto(intra_op_parallelism_threads=num_cores,inter_op_parallelism_threads=num_cores,
#    allow_soft_placement=True,device_count = {'CPU' : num_CPU, 'GPU' : num_GPU})
#session = tf.Session(config=config)
#K.set_session(session)
#os.environ['CUDA_VISIBLE_DEVICES']=K.tensorflow_backend._get_available_gpus()[0][-1]#use first GPUid


#or 
def getdims(x=(10000,200)):
    """
    return the dims for network
    """
    assert len(x)==2
    n_sample=x[0]
    if n_sample>20000:# may be need complex network
        dims=[x[-1],128,32]
    elif n_sample>10000:#10000
        dims=[x[-1],64,32]
    elif n_sample>5000: #5000
        dims=[x[-1],32,16] #16
    elif n_sample>2000:
        dims=[x[-1],128]
    elif n_sample>500:
        dims=[x[-1],64]
    else:
        dims=[x[-1],16]
    #dims=[x[-1],64,32] if n_sample>10000 else [x[-1],32,16]
    return dims



    
def train_single(data,dims=None,
        alpha=1.0,
        tol=0.005,
        init='glorot_uniform',
        n_clusters=None,
        louvain_resolution=1.0,
        n_neighbors=15,
        pretrain_epochs=300,
        batch_size=256,
        activation='relu',
        actincenter='tanh',
        drop_rate_SAE=0.2,
        is_stacked=True,
        use_earlyStop=True,
        use_ae_weights=False,
	save_encoder_weights=False,
        save_dir='result_tmp',
        max_iter=1000,
        epochs_fit=4, 
        num_Cores=20,
        use_GPU=True,
        random_seed=201809,
        verbose=True,
	do_tsne=False,
):
    if isinstance(data,AnnData):
        adata=data
    else:
        adata=sc.AnnData(data)
    #make sure dims 
    if dims is None:
        dims=getdims(adata.shape)
    assert dims[0]==adata.shape[-1],'the number of columns of x doesnot equal to the first element of dims, we must make sure that dims[0]==x.shape[0]'
    #if use_GPU and tf.test.is_gpu_available():
#set seed
    random.seed(random_seed)
    np.random.seed(random_seed)
    tf.set_random_seed(random_seed)
    import multiprocessing
    total_cpu=multiprocessing.cpu_count()
    num_Cores=int(num_Cores) if total_cpu>int(num_Cores) else int(math.ceil(total_cpu/2)) 
    print('The number of cpu in your computer is',total_cpu)
    if use_GPU:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        #os.environ['CUDA_VISIBLE_DEVICES']=K.tensorflow_backend._get_available_gpus()[0][-1]#use first GPUid
    else:
        #set only use cpu
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        K.set_session(tf.Session(config=tf.ConfigProto(intra_op_parallelism_threads=num_Cores, inter_op_parallelism_threads=num_Cores)))
    if not use_ae_weights and os.path.isfile(os.path.join(save_dir,"ae_weights.h5")):
        os.remove(os.path.join(save_dir,"ae_weights.h5"))
  
    tic=time()#recored time         
    desc=DESC_MODEL(dims=dims,
              x=adata.X,
              alpha=alpha,
              tol=tol,
              init=init,
              n_clusters=n_clusters,
              louvain_resolution=louvain_resolution,
              n_neighbors=n_neighbors,
              pretrain_epochs=pretrain_epochs,
              batch_size=batch_size,
              random_seed=random_seed,
              activation=activation,
              actincenter=actincenter,
              drop_rate_SAE=drop_rate_SAE,
              is_stacked=is_stacked,
              use_earlyStop=use_earlyStop,
              use_ae_weights=use_ae_weights,
	      save_encoder_weights=save_encoder_weights,
              save_dir=save_dir
    )
    desc.compile(optimizer=SGD(0.01,0.9),loss='kld')
    Embeded_z,q_pred=desc.fit(maxiter=max_iter,epochs_fit=epochs_fit)
    print("The desc has been trained successfully!!!!!!")
    if verbose:
        print("The summary of desc model is:")
        desc.model.summary()
    print("The runtime is ",time()-tic)
    y_pred=pd.Series(np.argmax(q_pred,axis=1),index=adata.obs.index,dtype='category')
    y_pred.cat.categories=list(range(len(y_pred.unique())))
    adata.obs['desc_'+str(louvain_resolution)]=y_pred
    adata.obsm['X_Embeded_z'+str(louvain_resolution)]=Embeded_z
    if do_tsne:
        sc.tl.tsne(adata,use_rep="X_Embeded_z"+str(louvain_resolution),learning_rate=150,n_jobs=num_Cores)
        adata.obsm["X_tsne"+str(louvain_resolution)]=adata.obsm["X_tsne"]
        sc.logging.msg(' tsne finished', t=True, end=' ', v=4)
        sc.logging.msg('and added\n'
                 '    \'X_tsne\''+str(louvain_resolution),'the tsne coordinates (adata.ob)\n', v=4)
    #prob_matrix
    adata.uns['prob_matrix'+str(louvain_resolution)]=q_pred
    return adata


def train(data,dims=None,
        alpha=1.0,
        tol=0.005,
        init='glorot_uniform',
        n_clusters=None,
        louvain_resolution=[0.8,1.0,1.2],
        n_neighbors=15,
        pretrain_epochs=300,
        batch_size=256,
        activation='relu',
        actincenter='tanh',
        drop_rate_SAE=0.2,
        is_stacked=True,
        use_earlyStop=True,
        use_ae_weights=True,
	save_encoder_weights=False,
        save_dir='result_tmp',
        max_iter=1000,
        epochs_fit=4, 
        num_Cores=20,
        use_GPU=True,
        random_seed=201809,
        verbose=True,
	do_tsne=False
): 
    """ Deep Embeded single cell clustering(DESC) API
    Conduct clustering for single cell data given in the anndata object or np.ndarray,sp.sparmatrix,or pandas.DataFrame
      
    
    Argument:
    ------------------------------------------------------------------
    data: :class:`~anndata.AnnData`, `np.ndarray`, `sp.spmatrix`,`pandas.DataFrame`
        The (annotated) data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    dims: `list`, the number of node in encoder layer, which include input dim, that is
    [1000,64,32] represents for the input dimension is 1000, the first hidden layer have 64 node, and second hidden layer(or bottle neck layers) have 16 nodes. if not specified, it will be decided automatically according to the sample size.
    
    alpha: `float`, optional. Default: `1.0`, the degree of t-distribution.
    tol: `float`, optional. Default: `0.005`, Stop criterion, clustering procedure will be stoped when the difference ratio betwen the current iteration and last iteration larger than tol.
    init: `str`,optional. Default: `glorot_uniform`.
        Initialization method used to initialize weights.

    n_clusters: `int`, optional. Default:`None`, if we specify it , the algorithm will use K-means initialize the cluster center after autoencoder model trained.
    louvain_resolution: `list  or str or float. like, louvain_resolution=1.2 or louvain_resolution=[0.2,0.4,0.8] or louvain_resolution="0.3,0.4,0.8" sep with ","
    n_neightbors, The size of local neighborhood (in terms of number of neighboring data points) used for connectivity matrix. Larger values result in more global views of the manifold, while smaller values result in more local data being preserved. In general values should be in the range 2 to 100. Lo 

    pretrain_epochs:'int',optional. Default:`300`,the number of epochs for autoencoder model. 

    batch_size: `int`, optional. Default:`256`, the batch size for autoencoder model and clustering model. 

    activation; `str`, optional. Default,`relu`. the activation function for autoencoder model,which can be 'elu,selu,softplus,tanh,siogmid et al.', for detail please refer to`https://keras.io/activations/`.

    actincenter: `str`, optional. Default,'tanh', the activation function for the last layer in encoder and decoder model.

    drop_rate_SAE=: `float`, optional. Default, `0.2`. The drop rate for Stacked autoencoder, which just for  finetuning. 

    is_stacked:`bool`,optional. Default,`True`.The model wiil be pretrained by stacked autoencoder if is_stacked==True.

    use_earlyStop:`bool`,optional. Default,`True`. Stops training if loss does not improve if given min_delta=1e-4, patience=10.

    use_ae_weights: `bool`, optional. Default, `True`. Whether use ae_weights that has been pretrained(which must saved in `save_dir/ae_weights.h5`)

    save_encoder_weights: `bool`, optional. Default, `False`, it will save inter_ae_weights for every 20 iterations. )

    save_dir: 'str',optional. Default,'result_tmp',some result will be saved in this directory.

    max_iter: `int`, optional. Default,`1000`. The maximum iteration for clustering.

    epochs_fit: `int`,optional. Default,`4`, updateing clustering probability for every epochs_fit.

    num_Cores: `int`, optional. Default,`20`. How many cpus use during tranning. if num_Cores > the max cpus in our computer , num_Cores will use  a half of cpus in your computer. 

    use_GPU=True, `bool`, optional. Default, `True`. it will use GPU to train model if GPU is avaliable 

    random_seed, `int`,optional. Default,`201809`. the random seed for random.seed,,,numpy.random.seed,tensorflow.set_random_seed

    verbose,`bool`, optional. Default, `True`. It will ouput the model summary if verbose==True.
    do_tsne,`bool`,optional. Default, `False`. Whethter do tsne for representation for each embededing
    ------------------------------------------------------------------
    """

    if isinstance(data,AnnData):
        adata=data
    elif isinstance(data,pd.DataFrame):
        adata=sc.AnnData(data,obs=data.index,var=data.columns)
    else:
        x=data.toarray() if issparse(data) else data
        adata=sc.AnnData(x)

    if dims is None:
        dims=getdims(adata.shape)

    if isinstance(louvain_resolution,float) or isinstance(louvain_resolution,int):
        louvain_resolution=[float(louvain_resolution)]
    elif  isinstance(louvain_resolution,str):
        louvain_resolution=list(map(float,louvain_resolution.split(",")))
    else:
        assert isinstance(louvain_resolution,list),'louvain_resolution must be either a string with spearator "," or a list like [1.0,2.0,3.0] '
        louvain_resolution=list(map(float,louvain_resolution))
    #
    time_start=time()
    for ith,resolution in enumerate(louvain_resolution):
        print("Start to process resolution=",str(resolution))
        use_ae_weights=use_ae_weights if ith==0 else True
        res=train_single(data=data,
            dims=dims,
            alpha=alpha,
            tol=tol,
            init=init,
            n_clusters=n_clusters,
            louvain_resolution=resolution,
            n_neighbors=n_neighbors,
            pretrain_epochs=pretrain_epochs,
            batch_size=batch_size,
            activation=activation,
            actincenter=actincenter,
            drop_rate_SAE=drop_rate_SAE,
            is_stacked=is_stacked,
            use_earlyStop=use_earlyStop,
            use_ae_weights=use_ae_weights,
	    save_encoder_weights=save_encoder_weights,
            save_dir=save_dir,
            max_iter=max_iter,
            epochs_fit=epochs_fit, 
            num_Cores=num_Cores,
            use_GPU=use_GPU,
            random_seed=random_seed,
            verbose=verbose,
	    do_tsne=do_tsne)
        #update adata
        data=res
    print("The run time for all resolution is ",time()-time_start)
    return data


if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='just for simple test train.py',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--use_GPU', default=True, type=bool)
    args = parser.parse_args()
    print(args)
    from load_mnist import load_mnist
    x,y=load_mnist(sample_size=10000,seed=0)
    print ('MNIST use ', x.shape)
    dims=[x.shape[-1],64,32]
    adata=train_single(x,dims,louvain_resolution=0.4,use_GPU=True)
    #adata=train(x,dims,louvain_resolution=[0.2,0.4],use_GPU=False)
    #adata2=train(x,dims,louvain_resolution='0.4,0.6,0.8',use_GPU=args.use_GPU,save_dir='mnist')
    #test for paul15()
    adata=sc.read("data/paul15/paul15.h5ad")
    sc.pp.filter_cells(adata,min_genes=10)
    sc.pp.normalize_per_cell(adata,counts_per_cell_after=1e4)
    sc.pp.log1p(adata)
    sc.pp.filter_genes(adata,min_cells=20)
    #adata=train(adata,louvain_resolution=0.4,use_GPU=True,use_ae_weights=False)
    #adata=train(adata,louvain_resolution="0.2,0.4,0.5",use_GPU=True,use_ae_weights=True,save_encoder_weights=True)
    #adata=train(adata,louvain_resolution='0.5,0.4,0.7',use_GPU=True,use_ae_weights=False,save_encoder_weights=True)

         
    
 



import os
os.environ['PYTHONHASHSEED'] = '0'
import keras
from keras.layers import Input, Dense, Dropout
from keras.models import Model, Sequential
from keras.optimizers import SGD
from keras.utils.vis_utils import plot_model
from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
import  math
import numpy as np
import random
import tensorflow as tf
#random.seed(201809)
#np.random.seed(201809)
#tf.set_random_seed(201809) if tf.__version__<="2.0" else tf.random.set_seed(201809)

class SAE(object):
    """ 
    Stacked autoencoders. It can be trained in layer-wise manner followed by end-to-end fine-tuning.
    For a 5-layer (including input layer) example:
        Autoendoers model: Input -> encoder_0->act -> encoder_1 -> decoder_1->act -> decoder_0;
        stack_0 model: Input->dropout -> encoder_0->act->dropout -> decoder_0;
        stack_1 model: encoder_0->act->dropout -> encoder_1->dropout -> decoder_1->act;
    
    Usage:
        from SAE import SAE
        sae = SAE(dims=[784, 500, 10])  # define a SAE with 5 layers
        sae.fit(x, epochs=100)
        features = sae.extract_feature(x)
        
    Arguments:
        dims: list of number of units in each layer of encoder. dims[0] is input dim, dims[-1] is units in hidden layer.
              The decoder is symmetric with encoder. So number of layers of the auto-encoder is 2*len(dims)-1
        act: activation (default='relu'), not applied to Input, Hidden and Output layers.
        drop_rate: drop ratio of Dropout for constructing denoising autoencoder 'stack_i' during layer-wise pretraining
        batch_size: `int`, optional. Default:`256`, the batch size for autoencoder model and clustering model.
        random_seed, `int`,optional. Default,`201809`. the random seed for random.seed,,,numpy.random.seed,tensorflow.set_random_seed
        actincenter: the activation function in last layer for encoder and last layer for encoder (avoiding the representation values and reconstruct outputs are all non-negative)
        init: `str`,optional. Default: `glorot_uniform`. Initialization method used to initialize weights.
        use_earlyStop: optional. Default,`True`. Stops training if loss does not improve if given min_delta=1e-4, patience=10.
        save_dir:'str',optional. Default,'result_tmp',some result will be saved in this directory.
    """
    def __init__(self, dims, act='relu', drop_rate=0.2, batch_size=32,random_seed=201809,actincenter="tanh",init="glorot_uniform",use_earlyStop=True,save_dir='result_tmp'): #act relu
        self.dims = dims
        self.n_stacks = len(dims) - 1
        self.n_layers = 2*self.n_stacks  # exclude input layer
        self.activation = act
        self.actincenter=actincenter #linear
        self.drop_rate = drop_rate
        self.init=init
        self.batch_size = batch_size
        #set random seed
        random.seed(random_seed)
        np.random.seed(random_seed)
        #tf.set_random_seed(random_seed)
        tf.set_random_seed(random_seed) if tf.__version__<"2.0" else tf.random.set_seed(random_seed)
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        #
        self.random_seed=random_seed
        self.use_earlyStop=use_earlyStop
        self.stacks = [self.make_stack(i,random_seed=self.random_seed+2*i) for i in range(self.n_stacks)]
        self.autoencoders ,self.encoder= self.make_autoencoders()
        plot_model(self.autoencoders, show_shapes=True, to_file=os.path.join(save_dir,'autoencoders.png'))
    def choose_init(self,init="glorot_uniform",seed=1):
        if init not in {'glorot_uniform','glorot_normal','he_normal','lecun_normal','he_uniform','lecun_uniform','RandomNormal','RandomUniform',"TruncatedNormal"}:
            raise ValueError('Invalid `init` argument: '
                             'expected on of {"glorot_uniform", "glorot_normal", "he_normal","he_uniform","lecun_normal","lecun_uniform","RandomNormal","RandomUniform","TruncatedNormal"} '
                             'but got', mode)
        if init=="glorot_uniform":
            res=keras.initializers.glorot_uniform(seed=seed)
        elif init=="glorot_normal":
            res=keras.initializers.glorot_normal(seed=seed)
        elif init=="he_normal":
            res=keras.initializers.he_normal(seed=seed)
        elif init=='he_uniform':
            res=keras.initializers.he_uniform(seed=seed)
        elif init=="lecun_normal":
            res=keras.initializer.lecun_normal(seed=seed)
        elif init=="lecun_uniform":
            res=keras.initializers.lecun_uniform(seed=seed)
        elif init=="RandomNormal":
            res=keras.initializers.RandomNormal(mean=0.0,stddev=0.04,seed=seed)
        elif init=="RandomUniform":
            res=keras.initializers.RandomUniform(minval=-0.05,maxval=0.05,seed=seed)
        else:
            res=keras.initializers.TruncatedNormal(mean=0.0, stddev=0.05, seed=seed)
        return res
        
        

    def make_autoencoders(self):
        """ Fully connected autoencoders model, symmetric.
        """
        # input
        x = Input(shape=(self.dims[0],), name='input')
        h = x

        # internal layers in encoder
        for i in range(self.n_stacks-1):
            h = Dense(self.dims[i + 1], kernel_initializer=self.choose_init(init=self.init,seed=self.random_seed+i),activation=self.activation, name='encoder_%d' % i)(h)

        # hidden layer,default activation is linear
        h = Dense(self.dims[-1],kernel_initializer=self.choose_init(init=self.init,seed=self.random_seed+self.n_stacks), name='encoder_%d' % (self.n_stacks - 1),activation=self.actincenter)(h)  # features are extracted from here

        y=h
        # internal layers in decoder       
        for i in range(self.n_stacks-1, 0, -1):
            y = Dense(self.dims[i], kernel_initializer=self.choose_init(init=self.init,seed=self.random_seed+self.n_stacks+i),activation=self.activation, name='decoder_%d' % i)(y)

        # output
        y = Dense(self.dims[0], kernel_initializer=self.choose_init(init=self.init,seed=self.random_seed+2*self.n_stacks),name='decoder_0',activation=self.actincenter)(y)

        return Model(inputs=x, outputs=y,name="AE"),Model(inputs=x,outputs=h,name="encoder")

    def make_stack(self, ith,random_seed=0):
        """ 
        Make the ith denoising autoencoder for layer-wise pretraining. It has single hidden layer. The input data is 
        corrupted by Dropout(drop_rate)
        
        Arguments:
            ith: int, in [0, self.n_stacks)
        """
        in_out_dim = self.dims[ith]
        hidden_dim = self.dims[ith+1]
        output_act = self.activation
        hidden_act = self.activation
        if ith == 0:
            output_act = self.actincenter# tanh, or linear
        if ith == self.n_stacks-1:
            hidden_act = self.actincenter #tanh, or linear
        model = Sequential()
        model.add(Dropout(self.drop_rate, input_shape=(in_out_dim,),seed=random_seed))
        model.add(Dense(units=hidden_dim, activation=hidden_act, kernel_initializer=self.choose_init(init=self.init,seed=random_seed),name='encoder_%d' % ith))
        model.add(Dropout(self.drop_rate,seed=random_seed+1))
        model.add(Dense(units=in_out_dim, activation=output_act,kernel_initializer=self.choose_init(init=self.init,seed=random_seed+1), name='decoder_%d' % ith))
        return model

    def pretrain_stacks(self, x, epochs=200,decaying_step=3):
        """ 
        Layer-wise pretraining. Each stack is trained for 'epochs' epochs using SGD with learning rate decaying 10
        times every 'epochs/3' epochs.
        
        Arguments:
            x: input data, shape=(n_samples, n_dims)
            epochs: epochs for each stack
            decayiing_step: learning rate multiplies 0.1 every 'epochs/decaying_step' epochs 
        """
        features = x
        for i in range(self.n_stacks):
            print( 'Pretraining the %dth layer...' % (i+1))
            for j in range(int(decaying_step)):  # learning rate multiplies 0.1 every 'epochs/decaying_step' epochs
                print ('learning rate =', pow(10, -1-j))
                self.stacks[i].compile(optimizer=SGD(pow(10, -1-j), momentum=0.9), loss='mse')
                if self.use_earlyStop is True:
                    callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=10,verbose=1,mode='auto')]
                    self.stacks[i].fit(features,features,callbacks=callbacks,batch_size=self.batch_size,epochs=math.ceil(epochs/decaying_step))
                else:
                    self.stacks[i].fit(x=features,y=features,batch_size=self.batch_size,epochs=math.ceil(epochs/decaying_step))
            print ('The %dth layer has been pretrained.' % (i+1))

            # update features to the inputs of the next layer
            feature_model = Model(inputs=self.stacks[i].input, outputs=self.stacks[i].get_layer('encoder_%d'%i).output)
            features = feature_model.predict(features)

    def pretrain_autoencoders(self, x, epochs=300):
        """
        Fine tune autoendoers end-to-end after layer-wise pretraining using 'pretrain_stacks()'
        Use SGD with learning rate = 0.1, decayed 10 times every 80 epochs
        
        Arguments:
        x: input data, shape=(n_samples, n_dims)
        epochs: training epochs
        """
        print ('Copying layer-wise pretrained weights to deep autoencoders')
        for i in range(self.n_stacks):
            name = 'encoder_%d' % i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())
            name = 'decoder_%d' % i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())

        print ('Fine-tuning autoencoder end-to-end')
        for j in range(math.ceil(epochs/50)):
            lr = pow(10, -j)
            print ('learning rate =', lr)
            self.autoencoders.compile(optimizer=SGD(lr, momentum=0.9), loss='mse')
            self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50)

    def fit(self, x, epochs=300,decaying_step=3): # use stacked autoencoder pretrain and fine tuning
        self.pretrain_stacks(x, epochs=int(epochs/2),decaying_step=decaying_step)
        self.pretrain_autoencoders(x, epochs=epochs)

    def fit2(self,x,epochs=300): #no stack directly tran 
        for j in range(math.ceil(epochs/50)):
            lr = pow(10, -j)
            print ('learning rate =', lr)
            self.autoencoders.compile(optimizer=SGD(lr, momentum=0.9), loss='mse')
            if self.use_earlyStop:
                callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=10,verbose=1,mode='auto')]
                self.autoencoders.fit(x=x,y=x,callbacks=callbacks,batch_size=self.batch_size,epochs=epochs)
            else:
                  self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50)

    def extract_feature(self, x):
        """
        Extract features from the middle layer of autoencoders(representation).
        
        Arguments:
        x: data
        """
        return self.encoder.predict(x)


if __name__ == "__main__":
    """
    An example for how to use SAE model on MNIST dataset. You can copy this file, and run `python3 SAE.py` in terminal
    """
    import numpy as np
    def load_mnist(sample_size=10000):
        # the data, shuffled and split between train and test sets
        from keras.datasets import mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x = np.concatenate((x_train, x_test))
        y = np.concatenate((y_train, y_test))
        x = x.reshape((x.shape[0], -1))
        print ('MNIST samples', x.shape)
        id0=np.random.choice(x.shape[0],sample_size,replace=False)
        return x[id0], y[id0]

    import os
    os.environ["CUDA_VISIBLE_DEVICES"]="-1" # no use GPU
    x,y=load_mnist(10000)
    db = 'mnist'
    n_clusters = 10
    # define and train SAE model
    sae = SAE(dims=[x.shape[-1], 64,32])
    sae.fit(x=x, epochs=400)
    sae.autoencoders.save_weights('weights_%s.h5' % db)

    # extract features
    print ('Finished training, extracting features using the trained SAE model')
    features = sae.extract_feature(x)
    print ('performing k-means clustering on the extracted features')
    from sklearn.cluster import KMeans
    km = KMeans(n_clusters, n_init=20)
    y_pred = km.fit_predict(features)
    from sklearn.metrics import normalized_mutual_info_score as nmi
    print ('K-means clustering result on extracted features: NMI =', nmi(y, y_pred))



import numpy as np
import pandas as pd
from scipy.sparse import issparse
from anndata import AnnData


def _log1p(x):
    if issparse(x):
        np.log1p(x.data, out=x.data)
    else:
        np.log1p(x, out=x)
    return x


def log1p(data, copy=False):
    if copy:
        data = data.copy()
    if isinstance(data, AnnData):
        _log1p(data.X)
    else:
        _log1p(data)
    return data if copy else None

def get_mean_var(X):
# - using sklearn.StandardScaler throws an error related to
#   int to long trafo for very large matrices
# - using X.multiply is slower
    if True:
        mean = X.mean(axis=0)
        if issparse(X):
            mean_sq = X.multiply(X).mean(axis=0)
            mean = mean.A1
            mean_sq = mean_sq.A1
        else:
            mean_sq = np.multiply(X, X).mean(axis=0)
        # enforece R convention (unbiased estimator) for variance
        var = (mean_sq - mean**2) * (X.shape[0]/(X.shape[0]-1))
    else:
        from sklearn.preprocessing import StandardScaler
        scaler = StandardScaler(with_mean=False).partial_fit(X)
        mean = scaler.mean_
        # enforce R convention (unbiased estimator)
        var = scaler.var_ * (X.shape[0]/(X.shape[0]-1))
    return mean, var


def scale_bygroup(adata,groupby,max_value=6):
    res=None
    assert isinstance(adata,AnnData),'adata must be AnnData class'
    adata.X=adata.X.toarray() if issparse(adata.X) else adata.X
    if groupby in adata.obs.keys():
        df=pd.Series(adata.obs[groupby],dtype="category")
        for category in df.cat.categories:
            tmp=adata[df==category]
            tmp=tmp.X
            tmp=np.asarray(tmp)
            mean0,var0=get_mean_var(tmp)
            sd0=np.sqrt(var0)
            sd0[sd0<=1e-5]=1e-5
            tmp-=mean0
            tmp/=sd0
            if max_value is not None:
                tmp[tmp>max_value]=max_value
            adata.X[df==category]=tmp.copy()
    else:
        print("Warning: The groupby:"+str(groupby)+ "you provided is not exists, we scale across all cells")
        res=adata.X
        #avoid all 0 columns
        mean0,var0=get_mean_var(res)
        sd0=np.sqrt(var0)
        sd0[sd0<=1e-5]=1e-5
        if issparse(res):
            res=res.toarray()
        res-=mean0
        res/=sd0
        res[res>max_value]=max_value
        adata.X=res
    return adata





from anndata import read_mtx
from anndata.utils import make_index_unique


import numpy as np

def _get_max_prob(prob_mat):
    index = np.argmax(prob_mat, axis=1)
    return prob_mat[np.array([range(prob_mat.shape[0])]), index][0].tolist()





from keras.layers import Input, Dense, Dropout,Activation,BatchNormalization
from keras.models import Model, Sequential 
from keras.optimizers import SGD
from keras.utils.vis_utils import plot_model
from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
import math
advanced_activations=('PReLU','LeakyRelU')

class SAE(object):
    """ 
    Stacked autoencoders. It can be trained in layer-wise manner followed by end-to-end fine-tuning.
    For a 5-layer (including input layer) example:
        Autoendoers model: Input -> encoder_0->act -> encoder_1 -> decoder_1->act -> decoder_0;
        stack_0 model: Input->dropout -> encoder_0->act->dropout -> decoder_0;
        stack_1 model: encoder_0->act->dropout -> encoder_1->dropout -> decoder_1->act;
    
    Usage:
        from SAE import SAE
        sae = SAE(dims=[784, 500, 10])  # define a SAE with 5 layers
        sae.fit(x, epochs=100)
        features = sae.extract_feature(x)
        
    Arguments:
        dims: list of number of units in each layer of encoder. dims[0] is input dim, dims[-1] is units in hidden layer.
              The decoder is symmetric with encoder. So number of layers of the auto-encoder is 2*len(dims)-1
        act: activation (default='relu'), not applied to Input, Hidden and Output layers.
        drop_rate: drop ratio of Dropout for constructing denoising autoencoder 'stack_i' during layer-wise pretraining
        batch_size: batch size
    """
    def __init__(self,
                dims,
                batchnorm=True,
                act='relu',
                actinlayercenter='tanh',
                drop_rate=0.2,
                batch_size=32,
                init="glorot_uniform",
                use_earlyStop=False,
                save_dir='SAE_result_tmp'
            ): 
        assert actinlayercenter in ["tanh","linear"]
        self.dims=dims
        self.n_stacks=len(dims)-1
        self.n_layers=2*self.n_stacks
        self.batchnorm=True
        self.act='relu'
        self.actinlayercenter='tanh'
        self.drop_rate=drop_rate
        self.batch_size=batch_size,
        self.init=init
        self.use_earlyStop=use_earlyStop
        self.stacks=[self.make_stack(i) for i in range(self.n_stacks)]
        self.make_autoencoders()
        plot_model(self.autoencoders,show_shapes=True,to_file="autoendoers.png")

    def make_autoencoders(self):
        """ Fully connected autoencoder model
        """
        x=Input(shape=(self.dims[0],),name='input')
        h=x
        #encoder layer
        for i in range(self.n_stacks-1):
            h=Dense(self.dims[i+1],kernel_initializer=self.init,activation=None,name='encoder_%d'%i)(h)
            if self.batchnorm:
                h=BatchNormalization(center=True, scale=False)(h)
            if self.act in advanced_activations:
                h=keras.layers.__dict__[self.act](name='encoder_%d'%i)(h)
            else:
                h=Activation(self.act,name="act_encoder_%d"%i)(h)
        #hidden layer, recommand to use  a linear or tanh activation 
        h=Dense(self.dims[-1],kernel_initializer=self.init,activation=None,name="encoder_%d"%(self.n_stacks-1))(h)
        if self.batchnorm:
            h=BatchNormalization(center=True, scale=False)(h)
        h=Activation(self.actinlayercenter,name="act_center")(h)
        #decoder layer
        for i in range(self.n_stacks-1,0,-1):
            h=Dense(self.dims[i],kernel_initializer=self.init,activation=None,name='decoder_%d'%i)(h)
            if self.batchnorm:
                h=BatchNormalization(center=True, scale=False)(h)
            if self.act in advanced_activations:
                h=keras.layers.__dict__[self.act](name='decoder_%d'%i)(h)
            else:
                h=Activation(self.act,name="act_decoder_%d"%i)(h)
        #output
        h=Dense(self.dims[0],kernel_initializer=self.init,name="decoder_0",activation=self.actinlayercenter)(h)
        #autoencodes
        self.autoencoders=Model(inputs=x,outputs=h,name="AE")
        #get encoder
        self.encoder=self.get_encoder()
         
    def get_encoder(self):
        name="encoder_"+str(self.n_stacks-1)
        ret= Model(inputs=self.autoencoders.input,
                   outputs=self.autoencoders.get_layer(name).output,name="encoder")
        return ret

    def make_stack(self,ith):
        """
            Make the ith autoencoder for layer-wise pretrainning. It has single hidden layer. The input data is corrupted by Dropout(drop_rate) if self.drop_rate>0
            Arugumets:
            ith: int type, in range(0,self.n_stacks)
        """
        in_out_dim=self.dims[ith]
        hidden_dim=self.dims[ith+1]
        in_act=self.actinlayercenter if ith==self.n_stacks-1 else self.act
        x=Input(shape=(in_out_dim,),name="input")
        h=x
        #encoder stack layer 
        if self.drop_rate>0.0:
            h = Dropout(self.drop_rate, name='encoder_dropout')(h)
        #h=Dense(units=hidden_dim,activation=None,kernel_initializer=self.init,kernel_regularizer=l1_l2(0.1,0.2,name='encoder_%d'%ith)
        h=Dense(units=hidden_dim,activation=None,kernel_initializer=self.init,name='encoder_%d'%ith)(h)
        if in_act in advanced_activations:
            h=keras.layers.__dict__[self.act](name='encoder_%d'%ith)(h)
        else:
            h=Activation(self.act,name='act_encoder_%d'%ith)(h)
        h=Dense(units=in_out_dim,activation=None,kernel_initializer=self.init,name='decoder_%d'%ith)(h)
        #decoder stack layer 
        if self.drop_rate>0.0:
            h = Dropout(self.drop_rate, name='decoder_dropout')(h)
        if in_act in advanced_activations:
            h=keras.layers.__dict__[self.act](name='decoder_%d'%ith)(h)
        else:
            h=Activation(self.act,name="act_decoder_%d"%ith)(h)
        model=Model(inputs=x,outputs=h)
        print("Heere",model,type(model))
        return model

    def pretrain_stacks(self,x,epochs=200,decaying_step=4):
        """ 
        Layer-wise pretraining. Each stack is trained for 'epochs' epochs using SGD with learning rate decaying 4
        times every 'epochs/4' epochs.
        
        Arguments:
            x: input data, shape=(n_samples, n_dims)
            decaying_step, default=4, 
            epochs: epochs for each stack
        """
        features=x
        print(type(features),features.shape)
        decaying_step=int(decaying_step)
        for i in range(self.n_stacks):
            #print('Pretraing the %dth layer...'%(i+1))
            print('Pretraing the %dth layer...'%i)
            for j in range(decaying_step):# learning rate multiple 0.1 every 'epochs/3' epochs
                print('learning rate=',pow(10,-1-j))
                self.stacks[i].compile(optimizer=SGD(pow(10,-1-j),momentum=0.9),loss='mse')
                if self.use_earlyStop is True:
                    print("sdfsd")
                    callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=10,verbose=1,mode='auto')]
                    self.stacks[i].fit(features,features,callbacks=callbacks,batch_size=self.batch_size,epochs=math.ceil(epochs/decaying_step)) 
                else:
                    print("sdfsd222")
                    print(self.stacks[i])
                    self.stacks[i].fit(x=features,y=features,batch_size=self.batch_size,epochs=math.ceil(epochs/4))
            print('The %dth has been pretrained successfully!!!!!'%(i+1))
            #update features to the inputs of the next stacked layer
            feature_model=Model(inputs=self.stacks[i].input, outputs=self.stacks[i].get_layer('encoder_%d'%i).output)
            features=feature_model.predict(features)
        print("All stacked autoencoder layers trained successfull!!!!")
    
    def pretrain_autoencoders(self,x,epochs=300):
        """
        Fine tune autoendoers end-to-end after layer-wise pretraining using 'pretrain_stacks()'
        Use SGD with learning rate = 0.1, decayed 10 times every 80 epochs
        
        Arguments: 
            :param x: input data, shape=(n_samples, n_dims)
            :param epochs: training epochs
        """
        print("Copy layer-wise pretrained weights to autoencoders")
        for i in range(self.n_stacks):
            name='encoder_%d'%i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())
            name="decoder_%d"%i
            self.autoencoders.get_layer(name).set_weights(self.stacks[i].get_layer(name).get_weights())
        print("Fine tunning autoendoer end to end!!")
        for j in range(math.ceil(epochs/50)):
            lr=pow(10,-j-1)
            print("Fine tuning autoencoders with learning rate="+str(lr))
            self.autoencoders.compile(optimizer=SGD(lr, momentum=0.9), loss='mse')
            #early stopping
            if self.use_earlyStop is True:
                callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=10,verbose=1,mode='auto')]
                self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size,callbacks=callbacks, epochs=50)
            else:
                self.autoencoders.fit(x=x, y=x, batch_size=self.batch_size, epochs=50)

    def fit(self,x,epochs=200):
        self.pretrain_stacks(x,epochs=epochs)
        self.pretrain_autoencoders(x,epochs=epochs)
            

if __name__=='__main__':
    """
    An example for how to use SAE model on MNIST dataset. In terminal run
            python SAE.py
    """
    import numpy as np
    def load_mnist():
        # the data, shuffled and split between train and test sets
        from keras.datasets import mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x = np.concatenate((x_train, x_test))
        y = np.concatenate((y_train, y_test))
        x = x.reshape((x.shape[0], -1))
        x = np.divide(x, 50.)  # normalize as it does in DEC paper
        print ('MNIST samples', x.shape)
        return x, y
    x,y=load_mnist()
    print("Test on x shape is ",x.shape)
    sae=SAE(dims=[x.shape[-1],64,32])
    sae.fit(x=x,epochs=150)
    #extract features
    features=sae.encoder.predict(x)
     



        

    

        
   			







from anndata import read_mtx
from anndata.utils import make_index_unique
import pandas as pd


def read_10X(data_path, var_names='gene_symbols'):

    adata = read_mtx(data_path + '/matrix.mtx').T

    genes = pd.read_csv(data_path + '/genes.tsv', header=None, sep='\t')
    adata.var['gene_ids'] = genes[0].values
    adata.var['gene_symbols'] = genes[1].values

    assert var_names == 'gene_symbols' or var_names == 'gene_ids', \
        'var_names must be "gene_symbols" or "gene_ids"'

    if var_names == 'gene_symbols':
        var_names = genes[1]
    else:
        var_names = genes[0]

    if not var_names.is_unique:
        var_names = make_index_unique(pd.Index(var_names))
        print('var_names are not unique, "make_index_unique" has applied')

    adata.var_names = var_names

    cells = pd.read_csv(data_path + '/barcodes.tsv', header=None, sep='\t')
    adata.obs['barcode'] = cells[0].values
    adata.obs_names = cells[0]
    return adata



"""
from MulticoreTSNE import MulticoreTSNE as mTSNE
from anndata import AnnData
import matplotlib.pyplot as plt

def run_tsne(
        data,
        return_tsne=True,
        tsne_plot=True,
        n_jobs=1,
        n_components=2,
        perplexity=40,
        learning_rate=200
        ):
    assert isinstance(data, AnnData), 'Required input should be an AnnData object from desc'
    assert hasattr(data.obsm, 'embedded'), 'The embedded matrix not found, run desc first'
    tsne = mTSNE(n_jobs=n_jobs, n_components=n_components, perplexity=perplexity, learning_rate=learning_rate). \
        fit_transform(data.obsm['embedded'])
    data.obsm['tsne'] = tsne

    if tsne_plot:
        plt.scatter(tsne[:, 0], tsne[:, 1], s=0.1, c=data.obs['ident'])
        plt.title("tSNE plot")
        plt.xlabel("tSNE_1")
        plt.ylabel("tSNE_2")

    return tsne if return_tsne else None
"""



# Copyright 2018 Xiangjie Li et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import division
import matplotlib
matplotlib.use('Agg')
import os,sys,argparse,textwrap
import scanpy.api as sc
import train as train

def parse_args():
    parser = argparse.ArgumentParser(description='Deep Embeded Single Cell Clustering!!')
    parser.add_argument('-i','--input',type=str,help=textwrap.dedent(''' Input is raw count data in TSV(tsv,tab,data)/CSV or .h5ad(anndata) format'
                         'Row/col names are mandatory. Note that TSV/CSV files must be in gene×cell layout where rows are genes and
                          cols are cells(scRNA-seq convention). Use the -t/--transpose option if your count matrix in cell×gene layou                          t.H5AD files must be in cell*gene format(stats and scanpy convention), or '
                          'if directory, "matrix.mtx","barcodes.tsv","genes.tsv" must be included and separate by "tab separator". 
                           matrix.mtx example: 
                            %%MatrixMarket matrix coordinate integer general
                            35635 29065 17570845
                            63 1 1
                            82 1 1
                            114 1 1
                            152 1 1
                            167 1 3
                            .....

                            barcodes.tsv example:
                            cellname    batchID nUMI ...
                            AAACATACAACCAC  tenx    20000   ....
                            AAACATTGAGCTAC  tenx    10000   ....
                            AAACATTGATCAGC  tenx    12000
                            AAACATTGATCAGC  smartseq    300000  ....
                            or when no columns name the first columns will be regarded as  cellname
                            .....
                    
                            genes.tsv example:
                            genename    geneID
                            MIR1302-10  ensembl1
                            FAM138A ensembl1
                            OR4F5   ensembl1
                            RP11-34P13.7    ensembl1
                            RP11-34P13.8    ensembl1
                            or when no columns name the first columns will be regarded as  genename, if only have two columns, 
                            the second column will be regarded as gene name" '''))
    parser.add_argument('-o','--output',type=str,default='adata.h5ad',help="The file name of the output`anndata`")
    #transpose if layout is gene*cell
    parser.add_argument('-t','--transpose',type=bool,default=False,help="Transpose input matrix(default:False) only True when input is cell*gene")
    #parameter to pre filter cells
    parser.add_argument('--prefilter_cells',type=bool,default=True,help='Prefilter cells')
    parser.add_argument('--c_min_genes',type=int,default=1,help='')    
    parser.add_argument('--c_max_genes',type=int,default=None,help='')    
    parser.add_argument('--c_min_counts',type=int,default=None,help='')    
    parser.add_argument('--c_max_counts',type=int,default=None,help='')    
    #parameter to filter cells according to mito_percent
    parser.add_argument('--filter_cells_mito',type=bool,default=True,help='Whether filter cells according to mito_percent') 
    parser.add_argument('--mito_percent_cutoff',type=float, default=0.10,help='The cell with mito_percent>0.20 will be filtered.')
    parser.add_argument('-nUMIbefore_filter_MT_gene',type=bool,default=True,help='Compute n_counts and n_gene before filtering MT-ge                        ne if nUMIbefore_filter_MT_gene==True')
    #parameter to normalization cell

    parser.add_argument('--normalize_per_cell',type=bool,default=False,help='Normalize cells. if normalize_cell=False')
    parser.add_argument('--counts_per_cell_after',default=1e4,action='store_true',help=textwrap.dedent(''' `float` or `None`, optional (default: 
            `1e4`)If `None`, after normalization, each cell has a total count equal  to the median of the 
             *counts_per_cell* before normalization.'''))
    parser.add_argument('--counts_per_cell',default=None,action='store_true',help='`float` or `None`, optional(default:`None`) Precomputed counts per cell.')
    #parameter to pre filter genes
    parser.add_argument('--prefilter_genes',type=bool,default=True,help='Prefilter cells')
    parser.add_argument('--g_min_cells',type=int,default=1,help='')    
    parser.add_argument('--g_max_cells',type=int,default=None,help='')    
    parser.add_argument('--g_min_counts',type=int,default=None,help='')    
    parser.add_argument('--g_max_counts',type=int,default=None,help='')  
    #parameter to find variable genes
    parser.add_argument('--find_variable_genes',type=bool,default=True,help='find variable genes, the parameter is same as scanpy.api.pp.filter_gene_dispersion')
    parser.add_argument('--flavor',type=str,default='seurat',help=textwrap.dedent(''' "{'seurat', 'cell_ranger'}, optional (default: 'seurat')
        Choose the flavor for computing normalized dispersion. If choosing
        'seurat', this expects non-logarithmized data - the logarithm of mean
        and dispersion is taken internally when `log` is at its default value
        `True`. For 'cell_ranger', this is usually called for logarithmized data
        - in this case you should set `log` to `False`. In their default
        workflows, Seurat passes the cutoffs whereas Cell Ranger passes
        `n_top_genes`.min_mean=0.0125, max_mean=3, min_disp=0.5, max_disp=`None` : `float`, optional
        If `n_top_genes` unequals `None`, these cutoffs for the means and the
        normalized dispersions are ignored." '''))
    parser.add_argument('--min_mean',type=float,default=0.0125)
    parser.add_argument('--max_mean',type=float,default=None)
    parser.add_argument('--min_disp',type=float,default=None)
    parser.add_argument('--max_disp',type=float,default=None)
    parser.add_argument('--n_top_genes',type=int,default=1000,help="Number of highly-variable genes to keep.")
    #parameter take lop1p
    parser.add_argument('--log1p',type=bool,default=True,help='Take log1p transformation after find variable genes')
    #parameter scale_by_group
    parser.add_argument('--scale_data',type=int,default=1,help='one of [0,1,2],0:no scale, 1:scale overall cells 2:scale by group')
    parser.add_argument('--group',type=str,default=None,help='Scale gene by group, group must be the columns of adata.')
    parser.add_argument('--max_value',type=str,default=None,help='The maxmum value after scale')
    
    #parameter desc model 
    parser.add_argument('--alpha',type=float,default=1.0,help='the degree of t-distribution')
    parser.add_argument('--tol',type=float,default=0.005,help='Stop criterion, clustering procedure will be stoped when the difference ratio betwen the current iteration and last iteration larger than tol')
    parser.add_argument('--init',type=str,default='glorot_uniform',help='Initializing distribution,detail refer to keras')
    parser.add_argument('--n_clusters',type=int,default=None,help="The number of clusters for K-means methods, if not sepcified, it will use louvain method to initialize cluster centroid")
    parser.add_argument('--louvain_resolution',type=float,default=[0.4],nargs='+',help="resolution for louvian method")
    parser.add_argument('--n_neighbors',type=int,default=15,help="The neighbors for connectivity matrix of cell")
    parser.add_argument('--pretrain_epochs',type=int,default=300,help='the number of epochs for autoencoder model')
    parser.add_argument('--batch_size',type=int,default=256,help='the batch size for autoencoder model and clustering model')
    parser.add_argument('--activation',type=str,default='relu',help="the activation function for autoencoder model,which can be `elu,selu,softplus,tanh,siogmid et al.`, for detail please refer to`https://keras.io/activations/")
    parser.add_argument('--actinlayer1',type=str,default='tanh',help=' the activation function for the last layer in encoder and decoder model')
    parser.add_argument('--drop_rate_SAE',type=float,default=0.2,help='The drop rate for Stacked autoencoder, which just for  finetuning')
    parser.add_argument('--is_stacked',type=bool,default=True,help='The model will be pretrained by stacked autoencoder if is_stacked==True')
    parser.add_argument('--use_earlyStop',type=bool,default=True,help='Stops training if loss does not improve if given min_delta=1e-4, patience=10.')
    parser.add_argument('--save_dir',type=str,default='result_tmp',help='The  directory for result')
    parser.add_argument('--max_iter',type=int,default=1000,help=' The maximum iteration for clustering')
    parser.add_argument('--epochs_fit',type=int,default=4,help='updateing clustering probability for every epochs_fit. If sample size is samll, we tend to set a larger value')
    parser.add_argument('--num_Cores',type=int,default=20,help='How many cpus use during tranning. if num_Cores > the max cpus in our computer , num_Cores will use  a half of cpus in your computer.')
    parser.add_argument('--use_GPU',type=bool,default=False,help='It will use GPU to train model if GPU is avaliable')
    parser.add_argument('--random_seed',type=int,default=201809,help='he random seed for random.seed,,,numpy.random.seed,tensorflow.set_random_seed')
    parser.add_argument('--verbose',type=bool,default=False,help='It will ouput the model summary if verbose==True.')
    return parser.parse_args()
 
def main():
    args=parse_args()
    print(args)
    import preprocessing as io
    try:
        import tensorflow as tf
    except ImportError:
        raise ImportError('DESC requires tensorflow. please follow instructions at https://www.tensorflow.org/install/ to install it')  
    #Step1. create data
    adata=io.read_dataset(args.input,transpose=args.transpose) 
    #Step2. prefilter data
    if args.prefilter_cells: 
        io.prefilter_cells(adata,min_counts=args.c_min_genes,max_counts=args.c_max_counts,min_genes=args.c_min_genes,max_genes=args.c_max_genes)
    #filter cells according mito_percent and filter MT_gene, and computer nUMI nGene before and after filter MT_genes
    if args.filter_cells_mito:
        io.filter_cells_mito(adata,mito_percent_cutoff=args.mito_percent_cutoff,nUMIbefore_filter_MT_gene=args.nUMIbefore_filter_MT_gene) 
    #normalize data
    if args.normalize_per_cell:
        sc.pp.normalize_per_cell(adata,counts_per_cell_after=args.counts_per_cell_after,counts_per_cell=args.counts_per_cell)
        adata.raw=sc.pp.log1p(adata,copy=True)

    ## prefilter genes
    if args.prefilter_genes: 
        io.prefilter_genes(adata,min_counts=args.g_min_counts,max_counts=args.g_min_counts,min_cells=args.g_min_cells,max_cells=args.g_max_cells)
    #find variable genes
    if args.find_variable_genes: 
        assert args.flavor in ["seurat","cell_ranger"],"flavor must be one of{'seurat', 'cell_ranger'}"
        sc.pp.filter_genes_dispersion(adata,max_mean=args.max_mean,min_disp=args.min_disp,max_disp=args.max_disp,n_top_genes=args.n_top_genes)
    #take log-transformation
    if args.log1p:
        sc.pp.log1p(adata)
    if args.scale_data:
        assert args.scale_data in [0,1,2],'`scale_data` must be of one of 0,1,2'
        adata=io.scale_by_group(adata,groupby=args.group,max_value=args.max_value)
    #from . import train
    dims=[adata.shape[-1],64,32]  
    adata=train.train(adata,
            alpha=args.alpha,
            tol=args.tol,
            init=args.init,
            n_clusters=args.n_clusters,
            louvain_resolution=args.louvain_resolution,
            n_neighbors=args.n_neighbors,
            pretrain_epochs=args.pretrain_epochs,
            batch_size=args.batch_size,
            activation=args.activation,
            actinlayer1=args.actinlayer1,
            drop_rate_SAE=args.drop_rate_SAE,
            is_stacked=args.is_stacked,
            use_earlyStop=args.use_earlyStop,
            save_dir=args.save_dir,
            max_iter=args.max_iter,
            epochs_fit=args.epochs_fit,
            num_Cores=args.num_Cores,
            use_GPU=args.use_GPU,
            random_seed=args.random_seed,
            verbose=args.verbose)
    print(adata)

if __name__=='__main__':
    main()



"""
Keras implementation for Deep Embedded Clustering (DEC) algorithm for single cell analysis:

        Xiangjie Li et al 

Usage:
    use `python DEC.py -h` for help.

Author:
    Xiangjie Li. 2018.5.8
"""
from __future__ import division
import matplotlib
matplotlib.use('Agg')# in order to save fig to local disk
import networkx as nx
import matplotlib.pyplot as plt
from time import time
import numpy as np
import keras.backend as K
from keras.engine.topology import Layer, InputSpec
from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
from keras.layers import Dense, Input
from keras.models import Model
from keras.optimizers import SGD
from keras import callbacks
from keras.initializers import VarianceScaling
from sklearn.cluster import KMeans
import scanpy.api as sc
import pandas as pd
from natsort import natsorted #call natsorted
import os
try:
    from .SAE import SAE  # load Stacked autoencoder
except:
    from SAE import SAE  # load Stacked autoencoder


class ClusteringLayer(Layer):
    """
    Clustering layer converts input sample (feature) to soft label, i.e. a vector that represents the probability of the
    sample belonging to each cluster. The probability is calculated with student's t-distribution.

    # Example
    ```
        model.add(ClusteringLayer(n_clusters=10))
    ```
    # Arguments
        n_clusters: number of clusters.
        weights: list of Numpy array with shape `(n_clusters, n_features)` witch represents the initial cluster centers.
        alpha: parameter in Student's t-distribution. Default to 1.0.
    # Input shape
        2D tensor with shape: `(n_samples, n_features)`.
    # Output shape
        2D tensor with shape: `(n_samples, n_clusters)`.
    """

    def __init__(self, n_clusters, weights=None, alpha=1.0, **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(ClusteringLayer, self).__init__(**kwargs)
        self.n_clusters = n_clusters
        self.alpha = alpha
        self.initial_weights = weights
        self.input_spec = InputSpec(ndim=2)

    def build(self, input_shape):
        assert len(input_shape) == 2
        input_dim = input_shape[1]
        self.input_spec = InputSpec(dtype=K.floatx(), shape=(None, input_dim))
        self.clusters = self.add_weight((self.n_clusters, input_dim), initializer='glorot_uniform', name='clusters')
        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights
        self.built = True

    def call(self, inputs, **kwargs):
        """ student t-distribution, as same as used in t-SNE algorithm.
                 q_ij = 1/(1+dist(x_i, u_j)^2), then normalize it.
        Arguments:
            inputs: the variable containing data, shape=(n_samples, n_features)
        Return:
            q: student's t-distribution, or soft labels for each sample. shape=(n_samples, n_clusters)
        """
        q = 1.0 / (1.0 + (K.sum(K.square(K.expand_dims(inputs, axis=1) - self.clusters), axis=2) / self.alpha))
        q **= (self.alpha + 1.0) / 2.0
        q = K.transpose(K.transpose(q) / K.sum(q, axis=1))
        return q

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 2
        return input_shape[0], self.n_clusters

    def get_config(self):
        config = {'n_clusters': self.n_clusters}
        base_config = super(ClusteringLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class DESC(object):
    def __init__(self,
                 dims,
                 x, # input matrix, row sample, col predictors 
                 alpha=1.0,
		 tol=0.005,
                 init='glorot_uniform', #initialization method
                 n_clusters=None,     # Number of Clusters, if provided, the clusters center will be initialized by K-means,
                 louvain_resolution=1.0, # resolution for louvain 
                 n_neighbors=15,    # the 
                 pretrain_epochs=300, # epoch for autoencoder
                 batch_size=256, #batch_size for autoencoder
		 activation='relu',
                 actinlayer1="tanh",# activation for the last layer in encoder, and first layer in the decoder 
                 drop_rate_SAE=0.2,
                 is_stacked=True,
                 use_earlyStop=True,
                 save_dir="result_tmp" # save result to save_dir, the default is "result". if recurvie path, there root dir must be exists, or there will be something wrong: for example : "/result_singlecell/dataset1" will return wrong if "result_singlecell" not exist
                 ):

        if not os.path.exists(save_dir):
            print("Create the directory:"+str(save_dir)+" to save result")
            os.mkdir(save_dir)
        self.dims = dims
        self.x=x #feature n*p, n:number of cells, p: number of genes
        self.alpha = alpha
        self.tol=tol
        self.init=init
        self.input_dim = dims[0]  # for clustering layer 
        self.n_stacks = len(self.dims) - 1
        self.is_stacked=is_stacked
        self.resolution=louvain_resolution
        self.n_neighbors=n_neighbors
        self.pretrain_epochs=pretrain_epochs
        self.batch_size=batch_size
        self.activation=activation
        self.actinlayer1=actinlayer1
        self.drop_rate_SAE=drop_rate_SAE
        self.is_stacked=is_stacked
        self.use_earlyStop=use_earlyStop
        self.save_dir=save_dir
        self.pretrain(n_clusters=n_clusters)

    def pretrain(self,n_clusters=None):
        sae=SAE(dims=self.dims,
				act=self.activation,
                drop_rate=self.drop_rate_SAE,
                batch_size=self.batch_size,
                actinlayer1=self.actinlayer1,
                init=self.init,
                use_earlyStop=self.use_earlyStop
           )
        # begin pretraining
        t0 = time()
        print("Checking whether %s  exists in the directory"%str(os.path.join(self.save_dir,'ae_weights,h5')))
        if not os.path.isfile(self.save_dir+"/ae_weights.h5"):
            if self.is_stacked:
                sae.fit(self.x,epochs=self.pretrain_epochs)
            else:
                sae.fit2(self.x,epochs=self.pretrain_epochs)
            self.autoencoder=sae.autoencoders
            self.encoder=sae.encoder
        else:
            sae.autoencoders.load_weights(os.path.join(self.save_dir,"ae_weights.h5"))
            self.autoencoder=sae.autoencoders
            self.encoder=sae.encoder
        print('Pretraining time is', time() - t0)
        #save ae results into disk
        if not os.path.isfile(os.path.join(self.save_dir,"ae_weights.h5")):
            self.autoencoder.save_weights(os.path.join(self.save_dir,'ae_weights.h5'))
            print('Pretrained weights are saved to %s /ae_weights.h5' % self.save_dir)
        #initialize cluster centers using louvain if n_clusters is not exist
        features=self.extract_features(self.x)
        features=np.asarray(features)
        if isinstance(n_clusters,int):
            print("...number of clusters have been specified, Initializing Cluster centroid  using K-Means")
            kmeans = KMeans(n_clusters=n_clusters, n_init=20)
            Y_pred_init = kmeans.fit_predict(features)
            self.init_pred= np.copy(Y_pred_init)
            self.n_clusters=n_clusters
            cluster_centers=kmeans.cluster_centers_
            self.init_centroid=[cluster_centers]
        else:
            print("...number of clusters is unknown, Initialize cluster centroid using louvain method")
            #can be replaced by other clustering methods
            #using louvain methods in scanpy
            adata=sc.AnnData(features)
            sc.pp.neighbors(adata, n_neighbors=self.n_neighbors)
            sc.tl.louvain(adata,resolution=self.resolution)
            Y_pred_init=adata.obs['louvain']
            self.init_pred=np.asarray(Y_pred_init,dtype=int)
            features=pd.DataFrame(features,index=np.arange(0,features.shape[0]))
            Group=pd.Series(self.init_pred,index=np.arange(0,features.shape[0]),name="Group")
            Mergefeature=pd.concat([features,Group],axis=1)
            cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
            self.n_clusters=cluster_centers.shape[0]
            self.init_centroid=[cluster_centers]
        #create desc clustering layer
        clustering_layer = ClusteringLayer(self.n_clusters,weights=self.init_centroid,name='clustering')(self.encoder.output)
        self.model = Model(inputs=self.encoder.input, outputs=clustering_layer)
        

    
    def load_weights(self, weights):  # load weights of DEC model
        self.model.load_weights(weights)

    def extract_features(self, x):
        return self.encoder.predict(x)

    def predict(self, x):  # predict cluster labels using the output of clustering layer
        q = self.model.predict(x, verbose=0)
        return q.argmax(1)

    @staticmethod
    def target_distribution(q):
        weight = q ** 2 / q.sum(0)
        return (weight.T / weight.sum(1)).T

    def compile(self, optimizer='sgd', loss='kld'):
        self.model.compile(optimizer=optimizer, loss=loss)
    
    def fit(self, maxiter=1e3, epochs_fit=5): # unsupervised
        save_dir=self.save_dir
        #step1 initial weights by louvain,or Kmeans
        self.model.get_layer(name='clustering').set_weights(self.init_centroid)
        # Step 2: deep clustering
        y_pred_last = np.copy(self.init_pred)
        for ite in range(int(maxiter)):
            q = self.model.predict(self.x, verbose=0)
            p = self.target_distribution(q)  # update the auxiliary target distribution p
            # evaluate the clustering performance
            y_pred = q.argmax(1)

             # check stop criterion
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
            y_pred_last = np.copy(y_pred)
            if ite > 0 and delta_label < self.tol:
                print('delta_label ', delta_label, '< tol ', self.tol)
                print('Reached tolerance threshold. Stop training.')
                break
            print("The value of delta_label of current",str(ite+1),"th iteration is",delta_label,">= tol",self.tol)
            #train on whole dataset on prespecified batch_size
            if self.use_earlyStop:
                callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=5,verbose=1,mode='auto')]
                self.model.fit(x=self.x,y=p,epochs=epochs_fit,batch_size=self.batch_size,callbacks=callbacks,shuffle=True,verbose=True)
            else:
                self.model.fit(x=self.x,y=p,epochs=epochs_fit,batch_size=self.batch_size,shuffle=True,verbose=True)
        y0=pd.Series(y_pred,dtype='category')
        y0.cat.categories=range(0,len(y0.cat.categories))
        print("The final prediction cluster is:")
        x=y0.value_counts()
        print(x.sort_index(ascending=True))
        Embeded_z=self.extract_features(self.x)
        return Embeded_z,q
         

if __name__ == "__main__":
    # setting the hyper parameters
    import argparse

    parser = argparse.ArgumentParser(description='DESC-class test',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--batch_size', default=256, type=int)
    parser.add_argument('--maxiter', default=2e4, type=int)
    parser.add_argument('--pretrain_epochs', default=100, type=int)
    parser.add_argument('--tol', default=0.005, type=float)
    parser.add_argument('--save_dir', default='results_tmp')
    args = parser.parse_args()
    print(args)
    import os
    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)

    # create mnist data to test 
    import numpy as np
    def load_mnist():
        # the data, shuffled and split between train and test sets
        from keras.datasets import mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x = np.concatenate((x_train, x_test))
        y = np.concatenate((y_train, y_test))
        x = x.reshape((x.shape[0], -1))
        x = np.divide(x, 50.)  # normalize as it does in DEC paper
        print ('MNIST samples', x.shape)
        return x, y

    x,y=load_mnist()
    sc.settings.figdir="./figures/mnist"
    init = 'glorot_uniform'
    #dims=[x.shape[-1], 500, 300, 100, 30]
    dims=[x.shape[-1],64,32]
    # prepare sample data to  the DESC model
    id=np.random.choice([True,False],size=x.shape[0],p=[0.3,0.7])
    x=x[id]
    y=y[id] 
    desc = DESC(dims=dims,x=x,louvain_resolution=0.3)
    desc.model.summary()
    t0 = time()
    desc.compile(optimizer=SGD(0.01, 0.9), loss='kld')
    Embeded_z,q_pred= desc.fit(maxiter=30,epochs_fit=4)
    y_pred=q_pred.max(axis=1)
    obs_info=pd.DataFrame()
    obs_info["y_true"]=pd.Series(y.astype("U"),dtype="category")
    obs_info["y_pred"]=pd.Series(y_pred.astype("U"),dtype="category")
    adata=sc.AnnData(x,obs=obs_info)
    adata.obsm["X_Embeded_z"]=Embeded_z
    print('clustering time: ', (time() - t0))
    
    




from ..datasets import pbmc_processed
from ..models.desc import train


def run_desc_test():
    print('Start to run a package test!')
    adata = pbmc_processed()
    adata = train(adata, dims=[100, 64, 16], louvain_resolution=0.1)
    return None



#from . import original as og
from . import tools
from . import models
from . import datasets
from . import utilities

from anndata import read_h5ad
from scanpy.api.pp import normalize_per_cell, highly_variable_genes, log1p, scale

from .models.desc import train
from .tools.test import run_desc_test
from .tools.read import read_10X
from .tools.write import write_desc_result
from .tools.preprocessing import scale_bygroup
#from .tools.downstream import run_tsne
__version__ = '1.0.0.6'






import scanpy.api as sc
import pandas as pd
import numpy as np
import scipy
import os
from scipy.sparse import issparse
import datetime 

def read_dataset(input_file,transpose=False):
    """
    Construct a anndata object
       
    """
    if os.path.isfile(input_file):
        print("The value os", os.path.isfile(input_file))
        if str(input_file).endswith('h5ad'):
            adata=sc.read(input_file)
        elif sum([str(input_file).endswith(str(i)) for i in ["tsv",'TSV','tab','data']]):
            adata=sc.read_text(input_file,sep="\t",first_column_names=True)
            if transpose:
                adata=adata.T
        elif sum([str(input_file).endswith(str(i)) for i in ['csv',"CSV"]]):
            adata=sc.read_text(input_file,sep=",",first_column_names=True) 
            if transpose:
                adata=adata.T
        else:
            #ValueError 'The file must be one of *.h5ad, *.tsv,*TSV,*.tab,*data, *csv,*CSV' 
            print("The file must be one of *.h5ad, *.tsv,*TSV,*.tab,*data, *csv,*CSV")
    else:
        #read folder
        mtx=sc.read_mtx(os.path.join(input_file,"matrix.mtx")) 
        num_lines = sum(1 for line in open(os.path.join(input_file,'barcodes.tsv')))
        cellinfo=pd.read_csv(os.path.join(input_file,"barcodes.tsv"),sep="\t",header=None if num_lines==mtx.shape[1] else 0)
        if not 'cellname' in cellinfo.columns:
            cellinfo['cellname']=cellinfo.iloc[:,0]
        num_lines = sum(1 for line in open(os.path.join(input_file,'genes.tsv')))
        geneinfo=pd.read_csv(os.path.join(input_file,"genes.tsv"),sep="\t",header=None if num_lines==mtx.shape[0] else 0)
        if not 'genename' in geneinfo.columns:
            geneinfo['genename']=geneinfo.iloc[:,1]# for 10x,the second columns is the genename, and the first column is gene_id
        #create anndata
        adata=sc.AnnData(mtx.X.T,obs=cellinfo,var=geneinfo)
        adata.obs_names=adata.obs["cellname"]
        adata.var_names=adata.var["genename"]
        adata.obs_names_make_unique(join="-")
        adata.var_names_make_unique(join="-")
    #create time
    now = datetime.datetime.now()
    adata.uns["ProjectName"]="DESC created in"+str(now.strftime("%Y-%m-%d %H:%M")) 
    print("Creat adata successfully! The adata infofation is", adata)
    return adata

        
def prefilter_cells(adata,min_counts=None,max_counts=None,min_genes=200,max_genes=None):
    if min_genes is None and min_counts is None and max_genes is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[0],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_genes=min_genes)[0]) if min_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_genes=max_genes)[0]) if max_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_obs(id_tmp)
    adata.raw=sc.pp.log1p(adata,copy=True) #check the rowname 
    print("the var_names of adata.raw: adata.raw.var_names.is_unique=:",adata.raw.var_names.is_unique)
   

    
        
def prefilter_genes(adata,min_counts=None,max_counts=None,min_cells=10,max_cells=None):
    if min_cells is None and min_counts is None and max_cells is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[1],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_cells=min_cells)[0]) if min_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_cells=max_cells)[0]) if max_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_var(id_tmp)
    sc.pp.filter_cells(adata,min_genes=1)#avoiding some cell all 0




def filter_cells_mito(adata, mito_percent_cutoff=0.10,nUMIbefore_filter_MT_gene=True):
    mito_genes=[name.startswith("MT-") for name in adata.var_names]
    if sum(mito_genes)>0:
        if issparse(adata.X):
            adata.obs['percent_mito'] = np.sum(adata[:, mito_genes].X, axis=1).A1 / np.sum(adata.X, axis=1).A1
        else:
            adata.obs['percent_mito']=np.sum(adata[:,mito_genes].X,axis=1)/np.sum(adata.X,axis=1)
    if 'percent_mito' in adata.obs.columns:
        if nUMIbefore_filter_MT_gene:
            if issparse(adata.X): 
                adata.obs['n_counts_before']=adata.X.sum(axis=1).A1
                adata.obs['n_genes_before']=(adata.X!=0).sum(axis=1).A1
            else:
                adata.obs['n_counts_before']=adata.X.sum(axis=1)
                adata.obs['n_genes_before']=(adata.X!=0).sum(axis=1)
            adata._inplace_subset_obs(adata.obs['percent_mito'] < mito_percent_cutoff)
            adata.inplace_subset_var(~np.asarray(mito_genes))
        else:
            adata._inplace_subset_var(~np.asarray(mito_genes))
            adata._inplace_subset_obs(adata.obs['percent_mito'] < mito_percent_cutoff)
            if issparse(adata.X): 
                adata.obs['n_counts_after']=adata.X.sum(axis=1).A1
                adata.obs['n_genes_after']=(adata.X!=0).sum(axis=1).A1
            else:
                adata.obs['n_counts_after']=adata.X.sum(axis=1)
                adata.obs['n_genes_after']=(adata.X!=0).sum(axis=1)
    else:
        if issparse(adata.X): 
            adata.obs['n_counts']=adata.X.sum(axis=1).A1
            adata.obs['n_genes']=(adata.X!=0).sum(axis=1).A1
        else:
            adata.obs['n_counts']=adata.X.sum(axis=1)
            adata.obs['n_genes']=(adata.X!=0).sum(axis=1)
        
def scale_by_group(adata,groupby,scale_data=0,max_value=6):
    if scale_data==0: return adata
    if scale_data==1: return sc.pp.scale(adata,zero_center=True, max_value=max_value,copy=True)
    if groupby in adata.obs.keys():
        df=pd.Series(adata.obs[groupby],dtype="category")
        for category in df.cat.categories:
            tmp=adata[df==category].X
            mean0,var0=get_mean_var(tmp)
            sd0=np.sqrt(var0)
            sd0[sd0<=1e-5]=1e-5
            if issparse(tmp):
                tmp=tmp.toarray()
            tmp-=mean0
            tmp/=sd0
            if max_value is not None:
                tmp[tmp>max_value]=max_value
            res=np.vstack((res,np.copy(tmp))) if res is not None else np.copy(tmp)
        adata.X=res
    else:
        adata=sc.pp.scale(adata,zero_center=True, max_value=max_value,copy=True)
    return adata

    

                      

def OriginalClustering(adata,resolution=1.2,n_neighbors=20,n_comps=50,n_PC=20,n_job=4,dotsne=True,doumap=True,dolouvain=True):
    #Do PCA directly
    sc.tl.pca(adata,n_comps=n_comps)
    n_pcs=n_PC if n_PC<n_comps else n_comps
    #Do tsne based pca result
    if dotsne:
        sc.tl.tsne(adata,random_state=2,learning_rate=150,n_pcs=n_PC,n_jobs=n_job)
        #Save original X
        adata.obsm["X_tsne.ori"]=adata.obsm['X_tsne']
    #Do umap 
    if doumap:
        sc.pp.neighbors(adata,n_neighbors=n_neighbors)
        sc.tl.umap(adata)
        #Save original
        adata.obsm['X_umap.ori']=adata.obsm['X_umap']
    if dolouvain:
        sc.tl.louvain(adata,resolution=resolution)
        adata.obs['louvain_ori']=adata.obs['louvain']
    print("OriginalClustering has completed!!!")

##
def first2prob(adata):
    first2ratio=[name for name in adata.uns.key() if str(name).startswith("prob_matrix")]
    for key_ in first2ratio:
        q_pred=adata.uns[key_]
        q_pred_sort=np.sort(q_pred,axis=1)
        y=q_pred_sort[:,-1]/q_pred_sort[:,-2]
        adata["first2ratio_"+str(key_).split("matrix")[1]]=y



#or 
def getdims(x=(10000,200)):
    """
    return the dims for network
    """
    assert len(x)==2
    n_sample=x[0]
    if n_sample>20000:# may be need complex network
        dims=[x[-1],128,32]
    elif n_sample>10000:#10000
        dims=[x[-1],64,32]
    elif n_sample>5000: #5000
        dims=[x[-1],32,16] #16
    elif n_sample>2000:
        dims=[x[-1],128]
    elif n_sample>500:
        dims=[x[-1],64]
    else:
        dims=[x[-1],16]
    #dims=[x[-1],64,32] if n_sample>10000 else [x[-1],32,16]
    return dims




from anndata import read_h5ad
from scanpy.api.pp import normalize_per_cell, highly_variable_genes, log1p, scale

from .test import run_desc_test
from .read import read_10X
from .write import write_desc_result
#from .downstream import run_tsne
#from .preprocessing import log1p











"""
Keras implement Deep learning enables accurate clustering and batch effect removal in single-cell RNA-seq analysis
"""
from __future__ import division
import os
import matplotlib
havedisplay = "DISPLAY" in os.environ
#if we have a display use a plotting backend
if havedisplay:
    matplotlib.use('TkAgg')
else:
    matplotlib.use('Agg')
os.environ['PYTHONHASHSEED'] = '0'
import networkx as nx
import matplotlib.pyplot as plt
from time import time as get_time
import numpy as np
import random
import tensorflow as tf
import keras.backend as K
from keras.engine.topology import Layer, InputSpec
from keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau,History
from keras.layers import Dense, Input
from keras.models import Model,load_model
from keras.optimizers import SGD
from keras import callbacks
from keras.initializers import VarianceScaling
from sklearn.cluster import KMeans
import scanpy.api as sc
import pandas as pd
from natsort import natsorted #call natsorted
import os
try:
    from .SAE import SAE  # this is for installing package
except:
    from SAE import SAE  #  this is for testing whether DescModel work or not 
random.seed(201809)
np.random.seed(201809)
tf.set_random_seed(201809) if tf.__version__<  "2.0" else tf.random.set_seed(201809)
#tf.set_random_seed(201809)


class ClusteringLayer(Layer):
    """
    Clustering layer converts input sample (feature) to soft label, i.e. a vector that represents the probability of the
    sample belonging to each cluster. The probability is calculated with student's t-distribution.

    # Example
    ```
        model.add(ClusteringLayer(n_clusters=10))
    ```
    # Arguments
        n_clusters: number of clusters.
        weights: list of Numpy array with shape `(n_clusters, n_features)` witch represents the initial cluster centers.
        alpha: parameter in Student's t-distribution. Default to 1.0.
    # Input shape
        2D tensor with shape: `(n_samples, n_features)`.
    # Output shape
        2D tensor with shape: `(n_samples, n_clusters)`.
    """

    def __init__(self, n_clusters, weights=None, alpha=1.0, **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(ClusteringLayer, self).__init__(**kwargs)
        self.n_clusters = n_clusters
        self.alpha = alpha
        self.initial_weights = weights
        self.input_spec = InputSpec(ndim=2)

    def build(self, input_shape):
        assert len(input_shape) == 2
        input_dim = input_shape[1]
        self.input_spec = InputSpec(dtype=K.floatx(), shape=(None, input_dim))
        self.clusters = self.add_weight((self.n_clusters, input_dim), initializer='glorot_uniform', name='clusters')
        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights
        self.built = True

    def call(self, inputs, **kwargs):
        """ student t-distribution, as same as used in t-SNE algorithm.
                 q_ij = 1/(1+dist(x_i, u_j)^2), then normalize it.
        Arguments:
            inputs: the variable containing data, shape=(n_samples, n_features)
        Return:
            q: student's t-distribution with degree alpha, or soft labels for each sample. shape=(n_samples, n_clusters)
        """
        q = 1.0 / (1.0 + (K.sum(K.square(K.expand_dims(inputs, axis=1) - self.clusters), axis=2) / self.alpha))
        q **= (self.alpha + 1.0) / 2.0
        q = K.transpose(K.transpose(q) / K.sum(q, axis=1))
        return q 

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 2
        return input_shape[0], self.n_clusters

    def get_config(self):
        config = {'n_clusters': self.n_clusters}
        base_config = super(ClusteringLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class ClusteringLayerGaussian(ClusteringLayer):
    def __init__(self, n_clusters, weights=None, alpha=1.0, **kwargs):
        super().__init__(n_clusters,weights,alpha,**kwargs)
    
    def call(self,inputs,**kwargs):
        sigma=1.0
        q=K.sum(K.exp(-K.square(K.expand_dims(inputs,axis=1)-self.clusters)/(2.0*sigma*sigma)),axis=2)
        q=K.transpose(K.transpose(q)/K.sum(q,axis=1))
        return q


class DescModel(object):
    def __init__(self,
                 dims,
                 x, # input matrix, row sample, col predictors 
                 alpha=1.0,
		 tol=0.005,
                 init='glorot_uniform', #initialization method
                 louvain_resolution=1.0, # resolution for louvain 
                 n_neighbors=10,    # the 
                 pretrain_epochs=300, # epoch for autoencoder
                 epochs_fit=4, #epochs for each update,int or float 
                 batch_size=256, #batch_size for autoencoder
                 random_seed=201809,
		 activation='relu',
                 actincenter="tanh",# activation for the last layer in encoder, and first layer in the decoder 
                 drop_rate_SAE=0.2,
                 is_stacked=True,
                 use_earlyStop=True,
                 use_ae_weights=False,
		 save_encoder_weights=False,
                 save_encoder_step=5,
                 save_dir="result_tmp",
                 kernel_clustering="t"
                 # save result to save_dir, the default is "result_tmp". if recurvie path, the root dir must be exists, or there will be something wrong: for example : "/result_singlecell/dataset1" will return wrong if "result_singlecell" not exist
                 ):

        if not os.path.exists(save_dir):
            print("Create the directory:"+str(save_dir)+" to save result")
            os.mkdir(save_dir)
        self.dims = dims
        self.x=x #feature n*p, n:number of cells, p: number of genes
        self.alpha = alpha
        self.tol=tol
        self.init=init
        self.input_dim = dims[0]  # for clustering layer 
        self.n_stacks = len(self.dims) - 1
        self.is_stacked=is_stacked
        self.resolution=louvain_resolution
        self.n_neighbors=n_neighbors
        self.pretrain_epochs=pretrain_epochs
        self.epochs_fit=epochs_fit
        self.batch_size=batch_size
        self.random_seed=random_seed
        self.activation=activation
        self.actincenter=actincenter
        self.drop_rate_SAE=drop_rate_SAE
        self.is_stacked=is_stacked
        self.use_earlyStop=use_earlyStop
        self.use_ae_weights=use_ae_weights
        self.save_encoder_weights=save_encoder_weights
        self.save_encoder_step=save_encoder_step
        self.save_dir=save_dir
        self.kernel_clustering=kernel_clustering
        #set random seed
        random.seed(random_seed)
        np.random.seed(random_seed)
        #tf.set_random_seed(random_seed)
        tf.set_random_seed(random_seed) if tf.__version__ < "2.0" else tf.random.set_seed(random_seed)
	#pretrain autoencoder
        self.pretrain()
        

    def pretrain(self):
        sae=SAE(dims=self.dims,
		act=self.activation,
                drop_rate=self.drop_rate_SAE,
                batch_size=self.batch_size,
                random_seed=self.random_seed,
                actincenter=self.actincenter,
                init=self.init,
                use_earlyStop=self.use_earlyStop,
                save_dir=self.save_dir
           )
        # begin pretraining
        t0 = get_time()
        if self.use_ae_weights: 
            print("Checking whether %s  exists in the directory"%str(os.path.join(self.save_dir,'ae_weights.h5')))
            if not os.path.isfile(self.save_dir+"/ae_weights.h5"):
                print("The file ae_weights.h5 is not exits")
                if self.is_stacked:
                    sae.fit(self.x,epochs=self.pretrain_epochs)
                else:
                    sae.fit2(self.x,epochs=self.pretrain_epochs)
                self.autoencoder=sae.autoencoders
                self.encoder=sae.encoder
            else:
                sae.autoencoders.load_weights(os.path.join(self.save_dir,"ae_weights.h5"))
                self.autoencoder=sae.autoencoders
                self.encoder=sae.encoder
        else:
            print("use_ae_weights=False, the program will rerun autoencoder")
            if self.is_stacked:
                sae.fit(self.x,epochs=self.pretrain_epochs)
            else:
                sae.fit2(self.x,epochs=self.pretrain_epochs)
            self.autoencoder=sae.autoencoders
            self.encoder=sae.encoder
        
        print('Pretraining time is', get_time() - t0)
        #save ae results into disk
        if not os.path.isfile(os.path.join(self.save_dir,"ae_weights.h5")):
            self.autoencoder.save_weights(os.path.join(self.save_dir,'ae_weights.h5'))
            self.encoder.save_weights(os.path.join(self.save_dir,'encoder_weights.h5'))
            print('Pretrained weights are saved to %s /ae_weights.h5' % self.save_dir)
        #save autoencoder model
        self.autoencoder.save(os.path.join(self.save_dir,"autoencoder_model.h5"))
        #initialize cluster centers using louvain if n_clusters is not exist
        features=self.extract_features(self.x)
        features=np.asarray(features)
        #if isinstance(n_clusters,int):
        if False:
            #saved for other initizlization methods in futher
            #print("...number of clusters have been specified, Initializing Cluster centroid  using K-Means")
            """
            kmeans = KMeans(n_clusters=n_clusters, n_init=20)
            Y_pred_init = kmeans.fit_predict(features)
            self.init_pred= np.copy(Y_pred_init)
            self.n_clusters=n_clusters
            cluster_centers=kmeans.cluster_centers_
            self.init_centroid=[cluster_centers]
            """
        else:
            print("...number of clusters is unknown, Initialize cluster centroid using louvain method")
            #can be replaced by other clustering methods
            #using louvain methods in scanpy
            adata0=sc.AnnData(features)
            if adata0.shape[0]>200000:
                np.random.seed(adata0.shape[0])#set  seed 
                adata0=adata0[np.random.choice(adata0.shape[0],200000,replace=False)] 
            sc.pp.neighbors(adata0, n_neighbors=self.n_neighbors,use_rep="X")
            sc.tl.louvain(adata0,resolution=self.resolution)
            Y_pred_init=adata0.obs['louvain']
            self.init_pred=np.asarray(Y_pred_init,dtype=int)
            if np.unique(self.init_pred).shape[0]<=1:
                #avoid only a cluster
                #print(np.unique(self.init_pred))
                exit("Error: There is only a cluster detected. The resolution:"+str(self.resolution)+"is too small, please choose a larger resolution!!")
            features=pd.DataFrame(adata0.X,index=np.arange(0,adata0.shape[0]))
            Group=pd.Series(self.init_pred,index=np.arange(0,adata0.shape[0]),name="Group")
            Mergefeature=pd.concat([features,Group],axis=1)
            cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
            self.n_clusters=cluster_centers.shape[0]
            self.init_centroid=[cluster_centers]
        #create desc clustering layer
        if self.kernel_clustering=="gaussian":
            clustering_layer = ClusteringLayerGaussian(self.n_clusters,weights=self.init_centroid,name='clustering')(self.encoder.output)
        else:
            clustering_layer = ClusteringLayer(self.n_clusters,weights=self.init_centroid,name='clustering')(self.encoder.output)
        self.model = Model(inputs=self.encoder.input, outputs=clustering_layer)
        

    
    def load_weights(self, weights):  # load weights of DEC model
        self.model.load_weights(weights)

    def extract_features(self, x):
        return self.encoder.predict(x)

    def predict(self, x):  # predict cluster labels using the output of clustering layer
        q = self.model.predict(x, verbose=0)
        return q.argmax(1)

    @staticmethod
    def target_distribution(q):
        weight = q ** 2 / q.sum(0)
        return (weight.T / weight.sum(1)).T

    def compile(self, optimizer='sgd', loss='kld'):
        self.model.compile(optimizer=optimizer, loss=loss)

    def fit_on_batch(self,maxiter=1e4,update_interval=200,save_encoder_step=4):
        save_dir=self.save_dir
        #step1 initial weights by louvain,or Kmeans
        self.model.get_layer(name='clustering').set_weights(self.init_centroid)
        # Step 2: deep clustering
        y_pred_last = np.copy(self.init_pred)
        index_array = np.arange(self.x.shape[0])
        index=0
        for ite in range(int(maxiter)):
            if self.save_encoder_weights and ite%(save_encoder_step*update_interval)==0:
                self.encoder.save_weights(os.path.join(self.save_dir,'encoder_weights_resolution_'+str(self.resolution)+"_"+str(ite)+'.h5'))
                print('Fine tuning encoder weights are saved to %s/encoder_weights.h5' % self.save_dir) 
            if ite % update_interval ==0:
                q=self.model.predict(self.x,verbose=0)
                p=self.target_distribution(q)
                y_pred=q.argmax(1)
                delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
                y_pred_last = np.copy(y_pred)
                print("The value of delta_label of current",str(ite+1),"th iteration is",delta_label,">= tol",self.tol)
                if ite > 0 and delta_label < self.tol:
                    print('delta_label ', delta_label, '< tol ', self.tol)
                    print('Reached tolerance threshold. Stop training.')
                    break
            idx=index_array[index * self.batch_size: min((index+1) * self.batch_size, self.x.shape[0])]
            loss = self.model.train_on_batch(x=self.x[idx], y=p[idx])
            index = index + 1 if (index + 1) * self.batch_size <= self.x.shape[0] else 0
        #save encoder model
        self.encoder.save(os.path.join(self.save_dir,"encoder_model.h5"))
        #load model
        #encoder=load_model("encoder.h5")
        #
        y0=pd.Series(y_pred,dtype='category')
        y0.cat.categories=range(0,len(y0.cat.categories))
        print("The final prediction cluster is:")
        x=y0.value_counts()
        print(x.sort_index(ascending=True))
        Embedded_z=self.extract_features(self.x)
        q=self.model.predict(self.x,verbose=0)
        return Embedded_z,q

             
    
    def fit_on_all(self, maxiter=1e3, epochs_fit=5,save_encoder_step=5): # unsupervised
        save_dir=self.save_dir
        #step1 initial weights by louvain,or Kmeans
        self.model.get_layer(name='clustering').set_weights(self.init_centroid)
        # Step 2: deep clustering
        y_pred_last = np.copy(self.init_pred)
        for ite in range(int(maxiter)):
            if self.save_encoder_weights and ite%save_encoder_step==0: #save ae_weights for every 5 iterations
                self.encoder.save_weights(os.path.join(self.save_dir,'encoder_weights_resolution_'+str(self.resolution)+"_"+str(ite)+'.h5'))
                print('Fine tuning encoder weights are saved to %s/encoder_weights.h5' % self.save_dir)
            q = self.model.predict(self.x, verbose=0)
            p = self.target_distribution(q)  # update the auxiliary target distribution p
            # evaluate the clustering performance
            y_pred = q.argmax(1)
             # check stop criterion
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
            y_pred_last = np.copy(y_pred)
            if ite > 0 and delta_label < self.tol:
                print('delta_label ', delta_label, '< tol ', self.tol)
                print('Reached tolerance threshold. Stop training.')
                break
            print("The value of delta_label of current",str(ite+1),"th iteration is",delta_label,">= tol",self.tol)
            #train on whole dataset on prespecified batch_size
            if self.use_earlyStop:
                callbacks=[EarlyStopping(monitor='loss',min_delta=1e-4,patience=5,verbose=1,mode='auto')]
                self.model.fit(x=self.x,y=p,epochs=epochs_fit,batch_size=self.batch_size,callbacks=callbacks,shuffle=True,verbose=True)
            else:
                self.model.fit(x=self.x,y=p,epochs=epochs_fit,batch_size=self.batch_size,shuffle=True,verbose=True)
        #save encoder model
        self.encoder.save(os.path.join(self.save_dir,"encoder_model.h5"))
        #load model
        #encoder=load_model("encoder.h5")
        #
        
        y0=pd.Series(y_pred,dtype='category')
        y0.cat.categories=range(0,len(y0.cat.categories))
        print("The final prediction cluster is:")
        x=y0.value_counts()
        print(x.sort_index(ascending=True))
        Embedded_z=self.extract_features(self.x)
        return Embedded_z,q

    def fit(self,maxiter=1e4):
        if isinstance(self.epochs_fit,int):
            embedded_z,q=self.fit_on_all(maxiter=maxiter,epochs_fit=self.epochs_fit,save_encoder_step=self.save_encoder_step)
        else:
            import math
            update_interval=math.ceil(self.epochs_fit*self.x.shape[0]/self.batch_size)
            embedded_z,q=self.fit_on_batch(maxiter=maxiter,save_encoder_step=self.save_encoder_step,update_interval=update_interval)
        return embedded_z,q
         

if __name__ == "__main__":
    # setting the hyper parameters
    import argparse

    parser = argparse.ArgumentParser(description='DescModel class test',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--batch_size', default=256, type=int)
    parser.add_argument('--maxiter', default=2e4, type=int)
    parser.add_argument('--pretrain_epochs', default=100, type=int)
    parser.add_argument('--tol', default=0.005, type=float)
    parser.add_argument('--save_dir', default='result_tmp')
    args = parser.parse_args()
    print(args)
    import os
    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)

    # create mnist data to test 
    import numpy as np
    def load_mnist(sample_size=10000):
        # the data, shuffled and split between train and test sets
        from keras.datasets import mnist
        (x_train, y_train), (x_test, y_test) = mnist.load_data()
        x = np.concatenate((x_train, x_test))
        y = np.concatenate((y_train, y_test))
        x = x.reshape((x.shape[0], -1))
        print ('MNIST samples', x.shape)
        id0=np.random.choice(x.shape[0],sample_size,replace=False)
        return x[id0], y[id0]
    #from load_mnist import load_mnist
    x,y=load_mnist(sample_size=10000)
    init = 'glorot_uniform'
    #dims=[x.shape[-1], 500, 300, 100, 30]
    dims=[x.shape[-1],64,32]
    # prepare sample data to  the DESC model
    import os
    os.environ["CUDA_VISIBLE_DEVICES"]="-1"
    desc = DescModel(dims=dims,x=x,louvain_resolution=0.8,use_ae_weights=True,epochs_fit=0.4)
    desc.model.summary()
    t0 = get_time()
    desc.compile(optimizer=SGD(0.01, 0.9), loss='kld')
    Embedded_z,q_pred= desc.fit(maxiter=30)
    y_pred=q_pred.max(axis=1)
    obs_info=pd.DataFrame()
    obs_info["y_true"]=pd.Series(y.astype("U"),dtype="category")
    obs_info["y_pred"]=pd.Series(y_pred.astype("U"),dtype="category")
    adata=sc.AnnData(x,obs=obs_info)
    adata.obsm["X_Embeded_z"]=Embedded_z
    print('clustering time: ', (get_time() - t0))
    
    



import os
from anndata import read_h5ad


def pbmc():
    adata = read_h5ad(os.path.dirname(__file__) + '/pbmc.h5ad')
    return adata


def pbmc_processed():
    adata = read_h5ad(os.path.dirname(__file__) + '/pbmc_processed.h5ad')
    return adata


def get_pbmc(save_dir='tmp_data'):
    if not os.path.isfile(save_dir + '/pbmc.h5ad'):
        adata = pbmc()
        adata.write(save_dir + '/pbmc.h5ad')
        print('The pbmc.h5ad has been put into ' + save_dir)
    else:
        print('The pbmc data has already exist!')
    return None
    


	


import numpy as np
#just for test desc work right
def load_mnist(sample_size=10000,seed=0):
    from keras.datasets import mnist
    np.random.seed(seed)
    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x = np.concatenate((x_train, x_test))
    y = np.concatenate((y_train, y_test))
    x = x.reshape((x.shape[0], -1))
    x = np.divide(x, 50.)  # normalize as it does in DEC paper
    id0=np.random.choice(x.shape[0],size=sample_size,replace=False) 
    return x[id0], y[id0]



import os
from  anndata import AnnData
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import scanpy.api as sc
from scipy.sparse import issparse

try:
    import tensorflow as tf
except ImportError:
    raise ImportError('DESC requires tensorflow  Please follow instructions'
                      'i at https://www.tensorflow.org/install/ to install'
                      ' it.')
#try:
#    from .train import train,train_single
#    from .load_mnist import load_mnist
#except:
from train import train, train_single
from load_mnist import load_mnist

def desc(data,
        dims=None,
        alpha=1.0,
        tol=0.005,
        init='glorot_uniform',
        n_clusters=None,
        louvain_resolution=1.0,
        n_neighbors=15,
        pretrain_epochs=300,
        batch_size=256,
        activation='relu',
        actinlayer1='tanh',
        drop_rate_SAE=0.2,
        is_stacked=True,
        use_earlyStop=True,
        save_dir='result_tmp',
        max_iter=1000,
        epochs_fit=4,
        num_Cores=20,
        use_GPU=True,
        random_seed=201809,
        verbose=True,
):
    """ Deep Embeded single cell clustering(DESC) API
    Conduct clustering for single cell data given in the anndata object or np.ndarray,sp.sparmatrix,or pandas.DataFrame
      
    
    Argument:
    ------------------------------------------------------------------
    data: :class:`~anndata.AnnData`, `np.ndarray`, `sp.spmatrix`,`pandas.DataFrame`
        The (annotated) data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    dims: `list`, the number of node in encoder layer, which include input dim, that is
    [1000,64,32] represents for the input dimension is 1000, the first hidden layer have 64 node, and second hidden layer(or bottle neck layers) have 16 nodes. if not specified, it will be decided automatically according to the sample size.
    
    alpha: `float`, optional. Default: `1.0`, the degree of t-distribution.
    tol: `float`, optional. Default: `0.005`, Stop criterion, clustering procedure will be stoped when the difference ratio betwen the current iteration and last iteration larger than tol.
    init: `str`,optional. Default: `glorot_uniform`.
        Initialization method used to initialize weights.

    n_clusters: `int`, optional. Default:`None`, if we specify it , the algorithm will use K-means initialize the cluster center after autoencoder model trained.
    louvain_resolution: `list  or str or float. like, louvain_resolution=1.2 or louvain_resolution=[0.2,0.4,0.8] or louvain_resolution="0.3,0.4,0.8" sep with ","
    n_neightbors, The size of local neighborhood (in terms of number of neighboring data points) used for connectivity matrix. Larger values result in more global views of the manifold, while smaller values result in more local data being preserved. In general values should be in the range 2 to 100. Lo 

    pretrain_epochs:'int',optional. Default:`300`,the number of epochs for autoencoder model. 

    batch_size: `int`, optional. Default:`256`, the batch size for autoencoder model and clustering model. 

    activation; `str`, optional. Default,`relu`. the activation function for autoencoder model,which can be 'elu,selu,softplus,tanh,siogmid et al.', for detail please refer to`https://keras.io/activations/`.

    actinlayer1: `str`, optional. Default,'tanh', the activation function for the last layer in encoder and decoder model.

    drop_rate_SAE=: `float`, optional. Default, `0.2`. The drop rate for Stacked autoencoder, which just for  finetuning. 

    is_stacked:`bool`,optional. Default,`True`.The model wiil be pretrained by stacked autoencoder if is_stacked==True.

    use_earlyStop=:`bool`,optional. Default,`True`. Stops training if loss does not improve if given min_delta=1e-4, patience=10.

    save_dir: 'str',optional. Default,'result_tmp',some result will be saved in this directory.

    max_iter: `int`, optional. Default,`1000`. The maximum iteration for clustering.

    epochs_fit: `int`,optional. Default,`4`, updateing clustering probability for every epochs_fit.

    num_Cores: `int`, optional. Default,`20`. How many cpus use during tranning. if num_Cores > the max cpus in our computer , num_Cores will use  a half of cpus in your computer. 

    use_GPU=True, `bool`, optional. Default, `True`. it will use GPU to train model if GPU is avaliable 

    random_seed, `int`,optional. Default,`201809`. the random seed for random.seed,,,numpy.random.seed,tensorflow.set_random_seed

    verbose,`bool`, optional. Default, `True`. It will ouput the model summary if verbose==True.
    ------------------------------------------------------------------
    """
    if isinstance(data,AnnData):
        adata=data.copy() 
    elif isinstance(data,pd.DataFrame):
        adata=sc.AnnData(data)
    else:
        x=data.toarray() if issparse(data) else data
        adata=sc.AnnData(x)
    if dims is None:
        dims=getdims(adata.shape) 
    if isinstance(louvain_resolution,float) or isinstance(louvain_resolution,int):
        adata=train_single(data=adata,
            dims=dims,
            alpha=alpha,
            tol=tol,
            init=init,
            n_clusters=n_clusters,
            louvain_resolution=float(louvain_resolution),
            n_neighbors=n_neighbors,
            pretrain_epochs=pretrain_epochs,
            batch_size=batch_size,
            activation=activation,
            actinlayer1=actinlayer1,
            drop_rate_SAE=drop_rate_SAE,
            is_stacked=is_stacked,
            use_earlyStop=use_earlyStop,
            save_dir=save_dir,
            max_iter=max_iter,
            epochs_fit=epochs_fit,
            num_Cores=num_Cores,
            use_GPU=use_GPU,
            random_seed=random_seed,
            verbose=verbose)
    else:
        adata=train(data=adata,
            dims=dims,
            alpha=alpha,
            tol=tol,
            init=init,
            n_clusters=n_clusters,
            louvain_resolution=louvain_resolution,
            n_neighbors=n_neighbors,
            pretrain_epochs=pretrain_epochs,
            batch_size=batch_size,
            activation=activation,
            actinlayer1=actinlayer1,
            drop_rate_SAE=drop_rate_SAE,
            is_stacked=is_stacked,
            use_earlyStop=use_earlyStop,
            save_dir=save_dir,
            max_iter=max_iter,
            epochs_fit=epochs_fit,
            num_Cores=num_Cores,
            use_GPU=use_GPU,
            random_seed=random_seed,
            verbose=verbose)
    return adata

#simple test
if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='just for simple test api.py',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--use_GPU', default=True, type=bool)
    args = parser.parse_args()
    print(args)
    x,_=load_mnist(sample_size=10000) 
    print('MNiST use',x.shape)
    dims=[x.shape[-1],64,32]
    adata1=desc(x,dims,louvain_resolution=0.4,use_GPU=args.use_GPU)
    adata2=desc(x,dims,louvain_resolution='0.4,0.6,0.8',use_GPU=args.use_GPU)
    print(adata1,adata2)

   
